//  crm_svm.c  - version v1.0
//
//  Copyright 2001-2006  William S. Yerazunis, all rights reserved.
//  
//  This software is licensed to the public under the Free Software
//  Foundation's GNU GPL, version 2.  You may obtain a copy of the
//  GPL by visiting the Free Software Foundations web site at
//  www.fsf.org, and a copy is included in this distribution.  
//
///////////////////////////////////////////////////////////////////////////
//
//    This code is originally copyright and owned by William
//    S. Yerazunis.  In return for addition of significant derivative
//    work, Yimin Wu is hereby granted a full unlimited license to use
//    this code, includng license to relicense under other licenses.
//
///////////////////////////////////////////////////////////////////////


//  Other licenses may be negotiated; contact the 
//  author for details.  
//
//  include some standard files
#include "crm114_sysincludes.h"
#include <math.h>

//  include any local crm114 configuration file
#include "crm114_config.h"

//  include the crm114 data structures file
#include "crm114_structs.h"

//  and include the routine declarations file
#include "crm114.h"


#include <stdlib.h>
#include <stdio.h>

//    the command line argc, argv
extern int prog_argc;
extern char **prog_argv;

//    the auxilliary input buffer (for WINDOW input)
extern char *newinputbuf;

//    the globals used when we need a big buffer  - allocated once, used 
//    wherever needed.  These are sized to the same size as the data window.
extern char *inbuf;
extern char *outbuf;
extern char *tempbuf;

//    The following sqrtf mumbojumbo because ppc_osx doesn't define sqrtf
//    like it should.
#ifndef sqrtf
#define sqrtf(x) sqrt((x))
#endif

#define LINEAR 0
#define C_SVC 0

typedef struct mythical_hyperspace_cell {
  unsigned long hash;
  // unsigned long key;
} HYPERSPACE_FEATUREBUCKET_STRUCT;
  


typedef struct mythical_svm_param {
  int svm_type;
  int kernel_type;
  double cache_size; // in MB
  double eps; //stop criteria
  double tau;
  double C; // parameter in C_SVC
  int shrinking; // use the shrinking heuristics
} SVM_PARAM;

typedef struct mythical_svm_problem {
  int l; //number of documents
  int *y; //label of documents -1/+1
  HYPERSPACE_FEATUREBUCKET_STRUCT **x; //x[i] is the ith document's feature vector
} SVM_PROBLEM;

typedef struct mythical_cache_node {
  struct mythical_cache_node *prev, *next;
  float *data;
  int len;
} CACHE_NODE;

typedef struct mythical_cache {
  int l; //the number of documents in the corpus
  long size; //the cache size (bytes)
  CACHE_NODE  *head;
  CACHE_NODE lru_headnode; //least-recent-use node
} CACHE;

typedef struct mythical_solver{
  double *alpha;
  double *G;  //Gradient of objective function
} SOLVER;

SVM_PARAM param;
SVM_PROBLEM svm_prob;
CACHE svmcache;
float *DiagQ;//diagonal Qmatrix
SOLVER solver;

static int hash_compare (void const *a, void const *b)
{
  HYPERSPACE_FEATUREBUCKET_STRUCT *pa, *pb;
  pa = (HYPERSPACE_FEATUREBUCKET_STRUCT *) a;
  pb = (HYPERSPACE_FEATUREBUCKET_STRUCT *) b;
  if (pa->hash < pb->hash)
    return (-1);
  if (pa->hash > pb->hash)
    return (1);
  return (0);
}

///////////////////////////////////////////////////////////////////////////
// 
//     Cache with least-recent-use strategy
//     This will be used to store part of Qmatrix
//

static void cache_init(int l, long size, CACHE *svmcache) {
  svmcache->l = l;
  svmcache->size = size;
  svmcache->head = (CACHE_NODE *)calloc(l, sizeof(CACHE_NODE));
  size /= sizeof(float);
  size -= l * (sizeof(CACHE_NODE)/sizeof(float));
  if(size < (2 * l)) 
    size = 2 * l; //cache size must at least  as large as two columns of Qmatrix
  (svmcache->lru_headnode).prev = (svmcache->lru_headnode).next = &(svmcache->lru_headnode);
}

static void cache_free(CACHE *svmcache){
  CACHE_NODE *temp;
  for(temp = (svmcache->lru_headnode).next; temp != &(svmcache->lru_headnode); temp = temp->next)
    free(temp->data);
  free(svmcache->head);
}

//delete one node
static void lru_delete(CACHE_NODE *h){
  h->prev->next = h->next;
  h->next->prev = h->prev;
}

//insert to the last positoin in the cache node list
static void lru_insert(CACHE_NODE *h, CACHE *svmcache){
  h->next = &(svmcache->lru_headnode);
  h->prev = (svmcache->lru_headnode).prev;
  h->prev->next = h;
  (svmcache->lru_headnode).prev = h;
}

//get data for certain document. return the length of cached data.If it is smaller than the request length, then we need to fill in the uncached data.
static int get_data(CACHE *svmcache, const int doc_index, float **data, int length){
  int result = length;
  CACHE_NODE *doc = svmcache->head + doc_index;
  if(doc->len) lru_delete(doc); //least-recent-use strategy
  
  //need to allocate more space
  if(length > (doc->len)){
    
    //cache hasn't enough free space, we need to release some old space
    while((svmcache->size) < (length - doc->len)){
      CACHE_NODE *temp = (svmcache->lru_headnode).next;
      lru_delete(temp);
      free(temp->data);
      svmcache->size += temp->len;
      temp->data = 0;
      temp->len = 0;
    }
    
    //allocate new space
    doc->data = (float *)realloc(doc->data, length * sizeof(float));
    svmcache->size -= (length - doc->len);
    result = doc->len;
    doc->len = length;
  }
  lru_insert(doc, svmcache);
  *data = doc->data;
  return result;
}


//dot operation of two feature vectors
static double dot(void const *a, void const *b)
{
  HYPERSPACE_FEATUREBUCKET_STRUCT *pa, *pb;
  pa = (HYPERSPACE_FEATUREBUCKET_STRUCT *) a;
  pb = (HYPERSPACE_FEATUREBUCKET_STRUCT *) b;
  int j = 0;
  //do{fprintf(stderr, "pa[%d]=%lud\n",j,pa[j].hash);}while(pa[j++].hash!=0);
  //j = 0;
  //do{fprintf(stderr, "pb[%d]=%lud\n",j,pb[j].hash);}while(pb[j++].hash!=0);
  int i = 0;
  j = 0;
  double sum = 0;
  while(pa[i].hash != 0 && pb[j].hash != 0)
  {
    if(pa[i].hash == pb[j].hash && pa[i].hash != 0){
      sum ++;
      i++;
      j++;
    }
    else{
      if(pa[i].hash > pb[j].hash)
	j++;
      else
	i++;
    }
  }
  //fprintf(stderr, "sum=%lf\n", sum);
  return sum;
}

//kernel operation of two feature vectors. Now only support linear kernel
static double kernel(void const *a, void const *b)
{
  switch(param.kernel_type){
  case LINEAR:
    return dot(a,b);
  default:
    return 0;
  }
}



//request of the ith column in Qmatrix 
static float *get_columnQ(int i,int length){
  float *columnQ;
  int uncached;
  if((uncached = get_data(&svmcache, i, &columnQ, length)) < length){
    int temp;
    for(temp = uncached; temp < length; temp++){
      columnQ[temp] = (float)svm_prob.y[i] * svm_prob.y[temp] * kernel(svm_prob.x[i],svm_prob.x[temp]);
    }
  }
  return columnQ;
}

//request of the diagonal in Qmatrix
static float *get_DiagQ(){
  float *DiagQ = (float *)malloc(svm_prob.l * sizeof(float));
  int i;
  for(i = 0; i<svm_prob.l; i++){
    DiagQ[i] = (float)svm_prob.y[i] * svm_prob.y[i] * kernel(svm_prob.x[i],svm_prob.x[i]);
  }
  return DiagQ;
}

//initialize the cache and diagonal Qmatrix
static void Q_init(){
  cache_init(svm_prob.l, param.cache_size*(1<<20), &svmcache);
  DiagQ = get_DiagQ();
}

// An SMO algorithm in Fan et al., JMLR 6(2005), p. 1889--1918
// Solves:
//
//	min 0.5(\alpha^T Q \alpha) + p^T \alpha
//
//		y^T \alpha = \delta
//		y_i = +1 or -1
//		0 <= alpha <= C
//	       
//
// Given:
//
//	Q, p, y, C, and an initial feasible point \alpha
//	l is the size of vectors and matrices
//	eps is the stopping tolerance
//
// solution will be put in \alpha, objective value will be put in obj
//

static void selectB(int workset[]){
  //select i
  int i = -1;
  double G_max = -HUGE_VAL;
  double G_min = HUGE_VAL;
  int t;
  for (t = 0; t < svm_prob.l; t++){
    if(((svm_prob.y[t] == 1) && (solver.alpha[t] < param.C)) || ((svm_prob.y[t] == -1) && (solver.alpha[t] > 0))){
      if(-svm_prob.y[t] * solver.G[t] >= G_max){
	i = t;
	G_max = -svm_prob.y[t] * solver.G[t];
      }
    }
  }
  //select j;
  int j = -1;
  double obj_min = HUGE_VAL;
  double a,b;
  float *Qi;
  for (t = 0; t< svm_prob.l; t++){
    if(((svm_prob.y[t] == -1) && (solver.alpha[t] < param.C)) || ((svm_prob.y[t] == 1) && (solver.alpha[t] > 0))){
      b = G_max + svm_prob.y[t] * solver.G[t];
      if(-svm_prob.y[t] * solver.G[t] <= G_min)
	G_min = -svm_prob.y[t] * solver.G[t];
      if(b > 0){
	if(i != -1){
	  Qi = get_columnQ(i,svm_prob.l);
	  a = Qi[i] + DiagQ[t] - 2 * svm_prob.y[i] * svm_prob.y[t] * Qi[t];
	  if (a <= 0)
	    a = param.tau;
	  if(-(b * b) / a <= obj_min){
	    j = t;
	    obj_min = -(b * b) / a;  
	  }
	}
	
      }
    }
  }
  if(G_max - G_min < param.eps){
    workset[0] = -1;
    workset[1] = -1;
  }else{
    workset[0] = i;
    workset[1] = j;
  }
}

//calculate objective value given alpha
static double calc_obj(){
  double obj = 0;
  int i;
  for (i=0; i < svm_prob.l; i++){
    obj += solver.alpha[i] * (solver.G[i] - 1);
  }
  obj /= 2;
  return obj;
}

static void solve(){
  int t,workset[2],i,j;
  double a,b, oldi, oldj, sum, obj = 0;
  float *Qi, *Qj;
  //initialize alpha to all zero;
  //initialize G to all -1;
  solver.alpha = (double *)malloc(svm_prob.l * sizeof(double));
  solver.G = (double *)malloc(svm_prob.l * sizeof(double));
  for(t = 0; t < svm_prob.l; t++){
    solver.alpha[t] = 0;
    solver.G[t] = -1;
  }
  while(1){
    selectB(workset);
    i = workset[0];
    j = workset[1];
    if(j == -1)
      break;
    Qi = get_columnQ(i, svm_prob.l);
    Qj = get_columnQ(j, svm_prob.l);
    
    a = Qi[i] + DiagQ[j] - 2 * svm_prob.y[i] * svm_prob.y[j] * Qi[j];
    if(a <= 0)
      a = param.tau;
    b = -svm_prob.y[i] * solver.G[i] + svm_prob.y[j] * solver.G[j];

    //update alpha
    oldi = solver.alpha[i];
    oldj = solver.alpha[j];
    solver.alpha[i] += svm_prob.y[i] * b/a;
    solver.alpha[j] -= svm_prob.y[j] * b/a;
    
    //project alpha back to the feasible region
    sum = svm_prob.y[i] * oldi + svm_prob.y[j] * oldj;
    if (solver.alpha[i] > param.C)
      solver.alpha[i] = param.C;
    if (solver.alpha[i] < 0 )
      solver.alpha[i] = 0;
    solver.alpha[j] = svm_prob.y[j] * (sum - svm_prob.y[i] * (solver.alpha[i]));
    if (solver.alpha[j] > param.C)
      solver.alpha[j] = param.C;
    if (solver.alpha[j] < 0 )
      solver.alpha[j] = 0;
    solver.alpha[i] = svm_prob.y[i] * (sum - svm_prob.y[j] * (solver.alpha[j]));
    
    
    //update gradient array
    for(t = 0; t < svm_prob.l; t++){  
      solver.G[t] += Qi[t] * (solver.alpha[i] - oldi) + Qj[t] * (solver.alpha[j] - oldj);
    }
    obj = calc_obj();
  }
}

//calculate b after achieving alpha
static double calc_b(){
  int count = 0;
  double upper = HUGE_VAL;
  double lower = -HUGE_VAL;
  double sum = 0;
  int i;
  double b;
  for(i = 0; i < svm_prob.l; i++){
    if(svm_prob.y[i] == 1){
      if(solver.alpha[i] == param.C){
	if(solver.G[i] > lower){
	  lower = solver.G[i];
	}
      }else if(solver.alpha[i] == 0){
	if(solver.G[i] < upper){
	  upper = solver.G[i];
	}
      }else{
	count++;
	sum += solver.G[i];
      }
	
    }else{
      if(solver.alpha[i] == 0){
	if(-solver.G[i] > lower){
	  lower = -solver.G[i];
	}
      }else if(solver.alpha[i] == param.C){
	if(-solver.G[i] < upper){
	  upper = -solver.G[i];
	}
      }else{
	count++;
	sum -= solver.G[i];
      }
    }
  }
  if(count > 0)
    b = -sum/count;
  else
    b = -(upper + lower)/2;
  return b;
}

//calculate the decision function
double calc_decision(HYPERSPACE_FEATUREBUCKET_STRUCT *x, double *alpha, double b){
  int i;
  double sum = 0;
  i=0;
  //int j = 0;
  //if(1){
  //do{fprintf(stderr, "x[%d][%d]=%lud\n",i,j,svm_prob.x[i][j].hash);}while(svm_prob.x[i][j++].hash!=0);  
  //}
  for (i = 0; i < svm_prob.l; i++){
    //fprintf(stderr,"x%d and x=%lf\n",i,kernel(x,svm_prob.x[i]));
    sum += svm_prob.y[i] * alpha[i] * kernel(x,svm_prob.x[i]);
  }
  sum += b;
  return sum;
}

//implementation of Lin's 2003 improved algorithm on Platt's probabilistic outputs for binary SVM
//input parameters: deci_array = array of svm decision values
//                  svm.prob
//                  posn = number of positive examples
//                  negn = number of negative examples

//outputs: parameters of sigmoid function-- A and B  (AB[0] = A, AB[1] = B)
void calc_AB(double *AB, double *deci_array, int posn, int negn){
  int maxiter = 100, i, j;
  double minstep = 1e-10;
  double sigma = 1e-3;
  double fval = 0.0;
  double hiTarget = (posn + 1.0) / (posn + 2.0);
  double loTarget = 1 / (negn + 2.0);
  double *t = (double *)malloc(svm_prob.l * sizeof(double));
  double fApB, h11, h22, h21, g1, g2, p, q, d1, d2, det, dA, dB, gd, stepsize, newA, newB, newf;
  for(i = 0; i< svm_prob.l; i++){
    if(svm_prob.y[i] > 0)
      t[i] = hiTarget;
    else
      t[i] = loTarget;
  }
  AB[0] = 0.0;
  AB[1] = log((negn + 1.0) / (posn + 1.0));
  for (i = 0; i < svm_prob.l; i++){
    fApB = deci_array[i] * AB[0] + AB[1];
    if(fApB >= 0)
      fval += t[i] * fApB + log(1 + exp(-fApB));
    else
      fval += (t[i] - 1) * fApB + log(1 + exp(fApB));
  }
  
  for(j = 0; j < maxiter; j++){
    h11 = h22 = sigma;
    h21 = g1 = g2 = 0.0;
    for(i = 0; i < svm_prob.l; i++){
      fApB = deci_array[i] * AB[0] + AB[1];
      if(fApB >= 0){
	p = exp(-fApB) / (1.0 + exp(-fApB));
	q = 1.0 / (1.0 + exp(-fApB));
      }else{
	p =  1.0 / (1.0 + exp(fApB));
	q = exp(fApB) / (1.0 + exp(fApB));
      }
      d2 = p * q;
      h11 += deci_array[i] * deci_array[i] * d2;
      h22 += d2;
      h21 += deci_array[i] * d2;
      d1 = t[i] - p;
      g1 += deci_array[i] * d1;
      g2 += d1; 
    }
    // Stopping Criteria
    if ((fabs(g1) < 1e-5) && (fabs(g2) < 1e-5)){
      //fprintf(stderr, "***g1=%lf\tg2=%lf\n",fabs(g1),fabs(g2));
      break;
    }
    //compute modified Newton directions
    det = h11 * h22 - h21 * h21;
    dA = -(h22 * g1 - h21 * g2) / det;
    dB = -(-h21 * g1 +  h11 * g2) / det;
    gd = g1 * dA + g2 * dB;
    stepsize = 1;
    while (stepsize >= minstep){
      newA = AB[0] + stepsize * dA;
      newB = AB[1] + stepsize * dB;
      newf = 0.0;
      for (i = 0; i < svm_prob.l; i++){
	fApB = deci_array[i] * newA + newB;
	if (fApB >= 0)
	  newf += t[i] * fApB + log(1 + exp(-fApB));
	else
	  newf += (t[i] - 1) * fApB + log(1 + exp(fApB));
      }
      // Check whether sufficient decrease is satisfied
      if (newf < fval + 0.0001 * stepsize * gd){
	AB[0] = newA;
	AB[1] = newB;
	fval = newf;
	break;
      }
      else
	stepsize /= 2.0;
    }
    if (stepsize < minstep){
      fprintf(stderr, "Line search fails in probability estimates\n");
      break;
    }
  }
  if (j >= maxiter)
    fprintf(stderr, "Reaching maximal iterations in  probability estimates\n");
  free(t);
}

double sigmoid_predict(double decision_value, double A, double B)
{
	double fApB = decision_value * A + B;
	if (fApB >= 0)
		return exp(-fApB) / (1.0 + exp(-fApB));
	else
		return 1.0 / (1 + exp(fApB)) ;
}


int crm_expr_svm_learn(CSL_CELL *csl, ARGPARSE_BLOCK *apb, 
		       char *txtptr, long txtstart, long txtlen){
  int yimin_trace = 0;
  internal_trace = 0;
  user_trace = 0;
  int i = 0, j = 0, k = 0;
  char ftext[MAX_PATTERN];
  long flen;
  char file1[MAX_PATTERN];
  char file2[MAX_PATTERN];
  char file3[MAX_PATTERN];
  FILE *hashf; 
  long textoffset;
  long textmaxoffset;
  HYPERSPACE_FEATUREBUCKET_STRUCT *hashes;  //  the hashes we'll sort 
  long hashcounts; 
  long cflags, eflags;
  long sense;
  long microgroom;
  long unique;
  long use_unigram_features;
  char ptext[MAX_PATTERN]; //the regrex pattern
  long plen;
  regex_t regcb;
  regmatch_t match[5]; 
  struct stat statbuf1;      //  for statting the hash file1
  struct stat statbuf2;      //  for statting the hash file2

  //            set our cflags, if needed.  The defaults are
  //            "case" and "affirm", (both zero valued).
  //            and "microgroom" disabled.
  cflags = REG_EXTENDED;
  eflags = 0;
  sense = +1;
  if (apb->sflags & CRM_NOCASE){
    cflags = cflags | REG_ICASE;
    eflags = 1;
    if (user_trace)
      fprintf (stderr, "turning oncase-insensitive match\n");
  };
  if (apb->sflags & CRM_REFUTE){
    sense = -sense;
    /////////////////////////////////////
    //    Take this out when we finally support refutation
    ////////////////////////////////////
    //      fprintf (stderr, "Hyperspace Refute is NOT SUPPORTED YET\n");
    //return (0);
    if (user_trace)
      fprintf (stderr, " refuting learning\n");
  };
  microgroom = 0;
  if (apb->sflags & CRM_MICROGROOM){
    microgroom = 1;
    if (user_trace)
      fprintf (stderr, " enabling microgrooming.\n");
  };
  unique = 0;
  if (apb->sflags & CRM_UNIQUE){
    unique = 1;
    if (user_trace)
      fprintf (stderr, " enabling uniqueifying features.\n");
  };

  use_unigram_features = 0;
  if (apb->sflags & CRM_UNIGRAM){
    use_unigram_features = 1;
    if (user_trace)
      fprintf (stderr, " using only unigram features.\n");
  };
   
  //   Note that during a LEARN in hyperspace, we do NOT use the mmap of
  //    pre-existing memory.  We just write to the end of the file instead.
  //    malloc up the unsorted hashbucket space 
  hashes = calloc (HYPERSPACE_MAX_FEATURE_COUNT, 
		   sizeof (HYPERSPACE_FEATUREBUCKET_STRUCT));
  hashcounts = 0;

  // extract the file names for storing svm solver.( file1.svm | file2.svm | 1vs2_solver.svm )
  crm_get_pgm_arg (ftext, MAX_PATTERN, apb->p1start, apb->p1len);
  flen = apb->p1len;
  flen = crm_nexpandvar (ftext, flen, MAX_PATTERN);
 
  strcpy(ptext,"[[:space:]]*([[:graph:]]*)[[:space:]]*\\|[[:space:]]*([[:graph:]]*)[[:space:]]*\\|[[:space:]]*([[:graph:]]*)[[:space:]]*");
  plen = strlen(ptext);   
  i = crm_regcomp (&regcb, ptext, plen, cflags);
  if ( i > 0){
    crm_regerror ( i, &regcb, tempbuf, data_window_size);
    nonfatalerror ("Regular Expression Compilation Problem:", tempbuf);
    goto regcomp_failed;
  };
  k = crm_regexec (&regcb, ftext,
		   flen, 5, match, 0, NULL);
  if( k==0 ){
    //get three input files.
    memmove(file1,&ftext[match[1].rm_so],(match[1].rm_eo-match[1].rm_so));
    file1[match[1].rm_eo-match[1].rm_so]='\000';
    memmove(file2,&ftext[match[2].rm_so],(match[2].rm_eo-match[2].rm_so));
    file2[match[2].rm_eo-match[2].rm_so]='\000';
    memmove(file3,&ftext[match[3].rm_so],(match[3].rm_eo-match[3].rm_so));
    file3[match[3].rm_eo-match[3].rm_so]='\000';
    if(internal_trace)
      fprintf(stderr, "file1=%s\tfile2=%s\tfile3=%s\n", file1, file2, file3);
  }else{
    if (ptext[0] != '\0') crm_regfree (&regcb);
    i = 0;
    while(ftext[i] < 0x021) i++;
    j = i;
    while(ftext[j] >= 0x021) j++;
    ftext[j] = '\000';
    strcpy(file1, &ftext[i]); 
    file2[0] = '\000';
    file3[0] = '\000';
  }
  //if (|Text|>0) hide it 
  
  //     get the "this is a word" regex
  crm_get_pgm_arg (ptext, MAX_PATTERN, apb->s1start, apb->s1len);
  plen = apb->s1len;
  plen = crm_nexpandvar (ptext, plen, MAX_PATTERN);
  
  //   compile the word regex
  //
  if ( internal_trace)
    fprintf (stderr, "\nWordmatch pattern is %s", ptext);

  i = crm_regcomp (&regcb, ptext, plen, cflags);
  
  if ( i > 0){
    crm_regerror ( i, &regcb, tempbuf, data_window_size);
    nonfatalerror ("Regular Expression Compilation Problem:", tempbuf);
    goto regcomp_failed;
  };
  
  //   Now tokenize the input text
  //   We got txtptr, txtstart, and txtlen from the caller.
  //
  textoffset = txtstart;
  textmaxoffset = txtstart + txtlen;
  i = 0;
  j = 0;
  k = 0;
  // if [Text]>0 hide it and append to the file1
  if (txtlen > 0){
    while (k == 0 && textoffset <= textmaxoffset 
	   && hashcounts < HYPERSPACE_MAX_FEATURE_COUNT  ){
      long wlen;
      long slen = textmaxoffset - textoffset;
      // if pattern is empty, extract non graph delimited tokens
      // directly ([[graph]]+) instead of calling regexec  (8% faster)
      if (ptext[0] != '\0'){
	k = crm_regexec (&regcb, &(txtptr[textoffset]),
			 slen, 5, match, 0, NULL);
      }else{
	k = 0;
	//         skip non-graphical characthers
	match[0].rm_so = 0;
	while (!isgraph (txtptr[textoffset + match[0].rm_so])
	       && textoffset + match[0].rm_so < textmaxoffset)
	  match[0].rm_so ++;
	match[0].rm_eo = match[0].rm_so;
	while (isgraph (txtptr [textoffset + match[0].rm_eo])
	       && textoffset + match[0].rm_eo < textmaxoffset)
	  match[0].rm_eo ++;
	if ( match[0].rm_so == match[0].rm_eo)
	  k = 1;
      };
      if (!(k != 0 || textoffset > textmaxoffset)){
	wlen = match[0].rm_eo - match[0].rm_so;
	
	memmove (tempbuf, 
		 &(txtptr[textoffset + match[0].rm_so]),
		 wlen);
	tempbuf[wlen] = '\000';
	if (internal_trace){
	  fprintf (stderr, 
		   "  Learn #%d t.o. %ld strt %ld end %ld len %ld is -%s-\n", 
		   i, 
		   textoffset,
		   (long) match[0].rm_so, 
		   (long) match[0].rm_eo,
		   wlen,
		   tempbuf);
	};
	if (match[0].rm_eo == 0){
	  nonfatalerror ( "The LEARN pattern matched zero length! ",
			  "\n Forcing an increment to avoid an infinite loop.");
	  match[0].rm_eo = 1;
	};
	
	//for now, only consider unigram
	unsigned long h1= strnhash (tempbuf, wlen);	
	//fprintf(stderr, "h1 = %ld\nhashcounts = %ld\n", h1,hashcounts);
	if (h1 == 0) h1 = 0xdeadbeef;
	hashes[hashcounts].hash = h1;
	hashcounts++;
	
	//  and account for the text used up.
	textoffset = textoffset + match[0].rm_eo;
	i++;
      } else{
	if (ptext[0] != '\0') crm_regfree (&regcb);
	k = 1;
      }
    };   //   end the while k==0
    
    //   Now sort the hashes array.
    //
    //hashcounts--;
    qsort (hashes, hashcounts, 
	   sizeof (HYPERSPACE_FEATUREBUCKET_STRUCT),
	   &hash_compare);
    
    if (user_trace){
      fprintf(stderr,"sorted hashes:\n");
      for(i=0;i<hashcounts;i++){
      fprintf(stderr, "hashes[%d]=%lud\n",i,hashes[i].hash);
      }
      fprintf (stderr, "Total hashes generated: %ld\n", hashcounts);
    }

    //   And uniqueify the hashes array
    //
    i = 0;
    j = 0;
    
    if (unique){
      while ( i < hashcounts ){
	if (hashes[i].hash != hashes[i+1].hash){
	hashes[j]= hashes[i];
	j++;
	};
	i++;
      };
      hashcounts = j;
    };
    
    //mark the end of a feature vector
    hashes[hashcounts].hash = 0;
 
    if (user_trace)
      fprintf (stderr, "Unique hashes generated: %ld\n", hashcounts);
    
    //append the hashed text to file1
    
    //  Because there are probably retained hashes of the 
    //  file, we need to force an unmap-by-name which will allow a remap
    //  with the new file length later on.
    crm_force_munmap_filename (file1);
    if (user_trace)
      fprintf (stderr, "Opening a svm file %s for append.\n", file1);
    hashf = fopen ( file1 , "ab+");
    if (user_trace)
      fprintf (stderr, "Writing to a svm file %s\n", file1);
    //    and write the sorted hashes out.
    fwrite (hashes, 1, 
	    sizeof (HYPERSPACE_FEATUREBUCKET_STRUCT) * (hashcounts + 1), 
	    hashf);
    fclose (hashf);
    
    //  let go of the hashes.
    free (hashes);
    //fprintf(stderr, "Finish hiding text to %s!\n", file1);
  }
  
  
  
  //if file2 is not empty, open file1 and file2, calculate hyperplane, 
  //and write the solution to file3
  if(file2[0] != '\000'){
    long file1_hashlens;
    HYPERSPACE_FEATUREBUCKET_STRUCT *file1_hashes;
    long file2_hashlens;
    HYPERSPACE_FEATUREBUCKET_STRUCT *file2_hashes;
    int k1, k2;
    k1 = stat (file1, &statbuf1);
    k2 = stat (file2, &statbuf2);
    if (k1 != 0) { 
      nonfatalerror ("Refuting from nonexistent data cannot be done!"
		     " More specifically, this data file doesn't exist: ",
		     file1);
      return (0);
    }else if(k2 != 0){
      nonfatalerror ("Refuting from nonexistent data cannot be done!"
		     " More specifically, this data file doesn't exist: ",
		     file2);
      return (0);
    }else{
      k1 = 0;
      k2 = 0;
      file1_hashlens = statbuf1.st_size;
      file1_hashes = (HYPERSPACE_FEATUREBUCKET_STRUCT *)
	crm_mmap_file (file1,
		       0, file1_hashlens,
		       PROT_READ | PROT_WRITE,
		       MAP_SHARED,
		       NULL);
      file1_hashlens = file1_hashlens 
	/ sizeof (HYPERSPACE_FEATUREBUCKET_STRUCT );
      file2_hashlens = statbuf2.st_size;
      file2_hashes = (HYPERSPACE_FEATUREBUCKET_STRUCT *)
	crm_mmap_file (file2,
		       0, file2_hashlens,
		       PROT_READ | PROT_WRITE,
		       MAP_SHARED,
		       NULL);
      file2_hashlens = file2_hashlens 
	/ sizeof (HYPERSPACE_FEATUREBUCKET_STRUCT );
      
      //find out how many documents in file1 and file2 separately
      for(i = 0;i< file1_hashlens;i++){
	if(yimin_trace)
	  fprintf (stderr, "\nThe %dth hash value in file1 is %lud", i, file1_hashes[i].hash);
	if(file1_hashes[i].hash == 0){
	  k1 ++;
	}
      }
      if(yimin_trace)
	fprintf (stderr, "\nThe total number of documents in file1 is %d\n", k1);
      
      for(i = 0;i< file2_hashlens;i++){
	if(yimin_trace)
	  fprintf (stderr, "\nThe %dth hash value in file2 is %lud", i, file2_hashes[i].hash);
	if(file2_hashes[i].hash == 0){
	  k2 ++;
	}
      }
      if(yimin_trace)
	fprintf (stderr, "\nThe total number of documents in file2 is %d\n", k2);
      
      if((k1 > 0) && (k2 >0)){
	//           extract parameters for svm 
	crm_get_pgm_arg(ptext, MAX_PATTERN, apb->s2start, apb->s2len);
	plen = apb->s2len;
	plen = crm_nexpandvar (ptext, plen, MAX_PATTERN);
	if(plen){
	  sscanf(ptext, "%d %d %lf %lf %lf %lf %d",&(param.svm_type), &(param.kernel_type), &(param.cache_size), &(param.eps), &(param.tau), &(param.C), &(param.shrinking));
	}else{
	  //set default parameters for SVM
	  param.svm_type = C_SVC;
	  param.kernel_type = LINEAR;
	  param.cache_size = 1;//MB
	  param.eps = 1e-3;
	  param.tau = 1e-12;
	  param.C = 1;
	  param.shrinking = 1;//not available right now
	}
	//initialize the svm_prob.x, svm_prob.y
	svm_prob.l = k1 + k2;
	int y[svm_prob.l];
	for(i = 0; i < k1; i++)
	  y[i] = 1;
	for(i = k1; i < svm_prob.l; i++)
	  y[i] = -1;
	svm_prob.y = y;
	HYPERSPACE_FEATUREBUCKET_STRUCT *x[svm_prob.l];
	x[0] = &(file1_hashes[0]);
	k = 1;
	for(i = 1;i< file1_hashlens - 1;i++){
	  if(file1_hashes[i].hash == 0){
	    x[k++] = &(file1_hashes[i+1]);
	  }
	}
	x[k++] = &(file2_hashes[0]);
	for(i = 1;i< file2_hashlens - 1;i++){
	  if(file2_hashes[i].hash == 0){
	    x[k++] = &(file2_hashes[i+1]);
	  }
	}
	svm_prob.x = x;
	if(yimin_trace){
	  for(i = 0;i< k;i++){
	    fprintf(stderr, "\nx[%d]=%lud\n",i,x[i][1].hash);
	  }       
	}
	Q_init();
	solve(); //result is in solver
	double b = calc_b();
	if(yimin_trace){
	   fprintf(stderr, "b=%lf\n",b);
	}
	//compute A,B for sigmoid prediction
	double *deci_array = (double*) malloc(svm_prob.l * sizeof(double));
	for(i = 0; i < svm_prob.l; i++){
	  deci_array[i] = calc_decision(svm_prob.x[i], solver.alpha, b);
	}
	double AB[2];
	calc_AB(AB, deci_array, k1, k2);
	//decision = sigmoid_predict(decision, AB[0], AB[1]);
	//free(hashes);
	free(deci_array);
	
	//  write solver to file3
	if (user_trace)
	  fprintf (stderr, "Opening a solution file %s for writing alpha and b.\n", file3);
	hashf = fopen ( file3 , "w+");
	if (user_trace)
	  fprintf (stderr, "Writing to a svm solution file %s\n", file3);

	
	fwrite(&k1, sizeof(int), 1, hashf);
	fwrite(&k2, sizeof(int), 1, hashf);
	
	//fprintf(hashf,"%d\t%d\n",k1,k2);
	for(i = 0; i < svm_prob.l; i++)
	  fwrite(&(solver.alpha[i]), sizeof(double), 1, hashf);
	  //fprintf(hashf,"%lf\t",solver.alpha[i]);
	//fprintf(hashf,"\n%lf",b);
	fwrite(&b, sizeof(double), 1, hashf);
	fwrite(&AB[0], sizeof(double), 1, hashf);
	fwrite(&AB[1], sizeof(double), 1, hashf);
	fclose (hashf);
        if(yimin_trace){
	  double decision = calc_decision(x[0],solver.alpha,b);
	  fprintf(stderr, "decision1 = %lf\n",decision);
	  decision = calc_decision(x[1],solver.alpha,b);
	  fprintf(stderr, "decision2 = %lf\n",decision);
	}
	//free cache
	cache_free(&svmcache);
	fprintf(stderr, "Finish calculating SVM hyperplane, store the solution to %s!\n", file3);
      }else{
	fprintf(stderr, "There hasn't enough documents to calculate a svm hyperplane!\n");
      }
      crm_force_munmap_filename (file1);
      crm_force_munmap_filename (file2);
      crm_force_munmap_filename (file3);
     
    }
    
    
  }//end if(file2[0])
 


  
 regcomp_failed:
  return 0;
}

int crm_expr_svm_classify(CSL_CELL *csl, ARGPARSE_BLOCK *apb,
			  char *txtptr, long txtstart, long txtlen){
  int yimin_trace = 0;
  int i,j, k;
  char ftext[MAX_PATTERN];
  long flen;
  char ptext[MAX_PATTERN]; //the regrex pattern
  long plen;
  char file1[MAX_PATTERN];
  char file2[MAX_PATTERN];
  char file3[MAX_PATTERN];
  regex_t regcb;
  regmatch_t match[5];
  FILE *hashf;
  long textoffset;
  long textmaxoffset;
  HYPERSPACE_FEATUREBUCKET_STRUCT *hashes;  //  the hashes we'll sort 
  long hashcounts; 
  long cflags, eflags;
  long sense;
  long microgroom;
  long unique;
  long use_unigram_features;
  struct stat statbuf1;      //  for statting the hash file1
  struct stat statbuf2;      //  for statting the hash file2
  struct stat statbuf3;      //  for statting the hash file3
  double *alpha;
  double b;
  double AB[2];
  long slen;
  char svrbl[MAX_PATTERN];  //  the match statistics text buffer
  long svlen;
  //  the match statistics variable
  char stext [MAX_PATTERN+MAX_CLASSIFIERS*(MAX_FILE_NAME_LEN+100)]; 
  long stext_maxlen = MAX_PATTERN+MAX_CLASSIFIERS*(MAX_FILE_NAME_LEN+100);
  double decision = 0;
  
  internal_trace = 0;
  user_trace = 0;

  //            extract the optional "match statistics" variable
  //
  crm_get_pgm_arg (svrbl, MAX_PATTERN, apb->p2start, apb->p2len);
  svlen = apb->p2len;
  svlen = crm_nexpandvar (svrbl, svlen, MAX_PATTERN);
  { 
    long vstart, vlen;
    crm_nextword (svrbl, svlen, 0, &vstart, &vlen);
    memmove (svrbl, &svrbl[vstart], vlen);
    svlen = vlen;
    svrbl[vlen] = '\000';
  };
  
  //     status variable's text (used for output stats)
  //    
  stext[0] = '\000';
  slen = 0;
  
  //            set our cflags, if needed.  The defaults are
  //            "case" and "affirm", (both zero valued).
  //            and "microgroom" disabled.
  cflags = REG_EXTENDED;
  eflags = 0;
  sense = +1;
  if (apb->sflags & CRM_NOCASE){
    cflags = cflags | REG_ICASE;
    eflags = 1;
    if (user_trace)
      fprintf (stderr, "turning oncase-insensitive match\n");
  };
  if (apb->sflags & CRM_REFUTE){
    sense = -sense;
    /////////////////////////////////////
    //    Take this out when we finally support refutation
    ////////////////////////////////////
    //      fprintf (stderr, "Hyperspace Refute is NOT SUPPORTED YET\n");
    //return (0);
    if (user_trace)
      fprintf (stderr, " refuting learning\n");
  };
  microgroom = 0;
  if (apb->sflags & CRM_MICROGROOM){
    microgroom = 1;
    if (user_trace)
      fprintf (stderr, " enabling microgrooming.\n");
  };
  unique = 0;
  if (apb->sflags & CRM_UNIQUE){
    unique = 1;
    if (user_trace)
      fprintf (stderr, " enabling uniqueifying features.\n");
  };

  use_unigram_features = 0;
  if (apb->sflags & CRM_UNIGRAM){
    use_unigram_features = 1;
    if (user_trace)
      fprintf (stderr, " using only unigram features.\n");
  };
  
  //   Note that during a LEARN in hyperspace, we do NOT use the mmap of
  //    pre-existing memory.  We just write to the end of the file instead.
  //    malloc up the unsorted hashbucket space 
  hashes = calloc (HYPERSPACE_MAX_FEATURE_COUNT, 
		   sizeof (HYPERSPACE_FEATUREBUCKET_STRUCT));
  hashcounts = 0;
  
  //     get the "this is a word" regex
  crm_get_pgm_arg (ptext, MAX_PATTERN, apb->s1start, apb->s1len);
  plen = apb->s1len;
  plen = crm_nexpandvar (ptext, plen, MAX_PATTERN);
  
  //   compile the word regex
  //
  if ( internal_trace)
    fprintf (stderr, "\nWordmatch pattern is %s", ptext);
  
  i = crm_regcomp (&regcb, ptext, plen, cflags);
  
  if ( i > 0){
    crm_regerror ( i, &regcb, tempbuf, data_window_size);
    nonfatalerror ("Regular Expression Compilation Problem:", tempbuf);
    goto regcomp_failed;
  };
  
  //   Now tokenize the input text
  //   We got txtptr, txtstart, and txtlen from the caller.
  //
  textoffset = txtstart;
  textmaxoffset = txtstart + txtlen;
  i = 0;
  j = 0;
  k = 0;

  //generate the sorted hashes of input text
  if(txtlen > 0){
    while (k == 0 && textoffset <= textmaxoffset 
	   && hashcounts < HYPERSPACE_MAX_FEATURE_COUNT  ){
      long wlen;
      long slen = textmaxoffset - textoffset;
      // if pattern is empty, extract non graph delimited tokens
      // directly ([[graph]]+) instead of calling regexec  (8% faster)
      if (ptext[0] != '\0'){
	k = crm_regexec (&regcb, &(txtptr[textoffset]),
			 slen, 5, match, 0, NULL);
      }else{
	k = 0;
	//         skip non-graphical characthers
	match[0].rm_so = 0;
	while (!isgraph (txtptr[textoffset + match[0].rm_so])
	       && textoffset + match[0].rm_so < textmaxoffset)
	  match[0].rm_so ++;
	match[0].rm_eo = match[0].rm_so;
	while (isgraph (txtptr [textoffset + match[0].rm_eo])
	       && textoffset + match[0].rm_eo < textmaxoffset)
	  match[0].rm_eo ++;
	if ( match[0].rm_so == match[0].rm_eo)
	  k = 1;
      };
      if (!(k != 0 || textoffset > textmaxoffset)){
	wlen = match[0].rm_eo - match[0].rm_so;
	
	memmove (tempbuf, 
		 &(txtptr[textoffset + match[0].rm_so]),
		 wlen);
	tempbuf[wlen] = '\000';
	if (internal_trace){
	  fprintf (stderr, 
		   "  Learn #%d t.o. %ld strt %ld end %ld len %ld is -%s-\n", 
		   i, 
		   textoffset,
		   (long) match[0].rm_so, 
		   (long) match[0].rm_eo,
		   wlen,
		   tempbuf);
	};
	if (match[0].rm_eo == 0){
	  nonfatalerror ( "The LEARN pattern matched zero length! ",
			  "\n Forcing an increment to avoid an infinite loop.");
	  match[0].rm_eo = 1;
	};
	
	//for now, only consider unigram
	unsigned long h1= strnhash (tempbuf, wlen);	
	//fprintf(stderr, "h1 = %lud\nhashcounts = %ld\n", h1,hashcounts);
	if (h1 == 0) h1 = 0xdeadbeef;
	hashes[hashcounts].hash = h1;
	hashcounts++;
	
	//  and account for the text used up.
	textoffset = textoffset + match[0].rm_eo;
	i++;
      } else{
	if (ptext[0] != '\0') crm_regfree (&regcb);
	k = 1;
      }
    };   //   end the while k==0
    //   Now sort the hashes array.
    //
    hashcounts--;
    qsort (hashes, hashcounts, 
	   sizeof (HYPERSPACE_FEATUREBUCKET_STRUCT),
	   &hash_compare);
    //mark the end of a feature vector
    hashes[++hashcounts].hash = 0;
    if (user_trace)
      fprintf (stderr, "Total hashes generated: %ld\n", hashcounts);
    
    //   And uniqueify the hashes array
    //
    i = 0;
    j = 0;
    
    if (unique){
      while ( i <= hashcounts ){
	if (hashes[i].hash != hashes[i+1].hash){
	  hashes[j]= hashes[i];
	  j++;
	};
	i++;
      };
      hashcounts = j;
    };
    if (user_trace)
      fprintf (stderr, "Unique hashes generated: %ld\n", hashcounts);
  }else{
    fprintf(stderr, "please input the text you want to classify! Try it again!\n");
    return 0;
  }
  
  // extract the file names.( file1.svm | file2.svm | 1vs2_solver.svm )
  crm_get_pgm_arg (ftext, MAX_PATTERN, apb->p1start, apb->p1len);
  flen = apb->p1len;
  flen = crm_nexpandvar (ftext, flen, MAX_PATTERN);
  
  strcpy(ptext,"[[:space:]]*([[:graph:]]*)[[:space:]]*\\|[[:space:]]*([[:graph:]]*)[[:space:]]*\\|[[:space:]]*([[:graph:]]*)[[:space:]]*");
  plen = strlen(ptext);   
  i = crm_regcomp (&regcb, ptext, plen, cflags);
  if ( i > 0){
    crm_regerror ( i, &regcb, tempbuf, data_window_size);
    nonfatalerror ("Regular Expression Compilation Problem:", tempbuf);
    goto regcomp_failed;
  };  
  k = crm_regexec (&regcb, ftext,
		   flen, 5, match, 0, NULL);
  if( k==0 ){
    //get three input files.
    memmove(file1,&ftext[match[1].rm_so],(match[1].rm_eo-match[1].rm_so));
    file1[match[1].rm_eo-match[1].rm_so]='\000';
    memmove(file2,&ftext[match[2].rm_so],(match[2].rm_eo-match[2].rm_so));
    file2[match[2].rm_eo-match[2].rm_so]='\000';
    memmove(file3,&ftext[match[3].rm_so],(match[3].rm_eo-match[3].rm_so));
    file3[match[3].rm_eo-match[3].rm_so]='\000';
    if(internal_trace)
      fprintf(stderr, "file1=%s\tfile2=%s\tfile3=%s\n", file1, file2, file3);
    
    //open all files,first check whether file3 is the current version solution.
    long file1_hashlens;
    HYPERSPACE_FEATUREBUCKET_STRUCT *file1_hashes;
    long file2_hashlens;
    HYPERSPACE_FEATUREBUCKET_STRUCT *file2_hashes;
    int k1, k2, k3;
    k1 = stat (file1, &statbuf1);
    k2 = stat (file2, &statbuf2);
    k3 = stat (file3, &statbuf3);
     
    if (k1 != 0) { 
      nonfatalerror ("Refuting from nonexistent data cannot be done!"
		     " More specifically, this data file doesn't exist: ",
		     file1);
      return (0);
    }else if(k2 != 0){
      nonfatalerror ("Refuting from nonexistent data cannot be done!"
		     " More specifically, this data file doesn't exist: ",
		     file2);
      return (0);
    }else{
      k1 = 0;
      k2 = 0;
      
      file1_hashlens = statbuf1.st_size;
 
      crm_force_munmap_filename (file1);
      crm_force_munmap_filename (file2);
    
      file1_hashes = (HYPERSPACE_FEATUREBUCKET_STRUCT *)
	crm_mmap_file (file1,
		       0, file1_hashlens,
		       PROT_READ | PROT_WRITE,
		       MAP_SHARED,
		       NULL);
      file1_hashlens = file1_hashlens 
	/ sizeof (HYPERSPACE_FEATUREBUCKET_STRUCT );
      file2_hashlens = statbuf2.st_size;
      file2_hashes = (HYPERSPACE_FEATUREBUCKET_STRUCT *)
	crm_mmap_file (file2,
		       0, file2_hashlens,
		       PROT_READ | PROT_WRITE,
		       MAP_SHARED,
		       NULL);
      file2_hashlens = file2_hashlens 
	/ sizeof (HYPERSPACE_FEATUREBUCKET_STRUCT );
      
      //find out how many documents in file1 and file2 separately
      for(i = 0;i< file1_hashlens;i++){
	if(yimin_trace)
	  fprintf (stderr, "\nThe %dth hash value in file1 is %lud", i, file1_hashes[i].hash);
	if(file1_hashes[i].hash == 0){
	  k1 ++;
	}
      }
      if(yimin_trace)
	fprintf (stderr, "\nThe total number of documents in file1 is %d\n", k1);
      
      for(i = 0;i< file2_hashlens;i++){
	if(yimin_trace)
	  fprintf (stderr, "\nThe %dth hash value in file2 is %lud", i, file2_hashes[i].hash);
	if(file2_hashes[i].hash == 0){
	  k2 ++;
	}
      }
      if(yimin_trace)
	fprintf (stderr, "\nThe total number of documents in file2 is %d\n", k2);
      hashf = fopen ( file3 , "r+");
      int temp_k1, temp_k2;
      if(k3 == 0){ 
	//hashf = fopen ( file3 , "r+");
	fread(&temp_k1, sizeof(int), 1, hashf);
	fread(&temp_k2, sizeof(int), 1, hashf);
	//fscanf(hashf,"%d\t%d", &temp_k1, &temp_k2);
	if (yimin_trace)
	  fprintf(stderr, "temp_k1=%d\ttemp_k2=%d\n",temp_k1,temp_k2);
      }
      //assign svm_prob.x, svm_prob.y
      svm_prob.l = k1 + k2;
      int y[svm_prob.l];
      for(i = 0; i < k1; i++)
	y[i] = 1;
      for(i = k1; i < svm_prob.l; i++)
	y[i] = -1;
      svm_prob.y = y;
      HYPERSPACE_FEATUREBUCKET_STRUCT *x[svm_prob.l];
      x[0] = &(file1_hashes[0]);
      k = 1;
      for(i = 1;i< file1_hashlens - 1;i++){
	if(file1_hashes[i].hash == 0){
	  x[k++] = &(file1_hashes[i+1]);
	}
      }
      x[k++] = &(file2_hashes[0]);
      for(i = 1;i< file2_hashlens - 1;i++){
	if(file2_hashes[i].hash == 0){
	  x[k++] = &(file2_hashes[i+1]);
	}
      }
      svm_prob.x = x;
      if((k3 != 0) || (temp_k1 != k1) || (temp_k2 != k2)){
	fprintf(stderr, "SVM solution file is not up-to-date! we'll recalculate it!\n");
	//recalculate the svm solution
	if((k1 > 0) && (k2 >0)){
	  //           extract parameters for svm 
	  crm_get_pgm_arg(ptext, MAX_PATTERN, apb->s2start, apb->s2len);
	  plen = apb->s2len;
	  plen = crm_nexpandvar (ptext, plen, MAX_PATTERN);
	  if(plen){
	    sscanf(ptext, "%d %d %lf %lf %lf %lf %d",&(param.svm_type), &(param.kernel_type), &(param.cache_size), &(param.eps), &(param.tau), &(param.C), &(param.shrinking));
	  }else{
	    //set default parameters for SVM
	    param.svm_type = C_SVC;
	    param.kernel_type = LINEAR;
	    param.cache_size = 1;//MB
	    param.eps = 1e-3;
	    param.tau = 1e-12;
	    param.C = 1;
	    param.shrinking = 1;//not available right now
	  }
	  #ifdef yimin
	  //initialize the svm_prob.x, svm_prob.y
	  svm_prob.l = k1 + k2;
	  int y[svm_prob.l];
	  for(i = 0; i < k1; i++)
	    y[i] = 1;
	  for(i = k1; i < svm_prob.l; i++)
	    y[i] = -1;
	  svm_prob.y = y;
	  HYPERSPACE_FEATUREBUCKET_STRUCT *x[svm_prob.l];
	  x[0] = &(file1_hashes[0]);
	  k = 1;
	  for(i = 1;i< file1_hashlens - 1;i++){
	    if(file1_hashes[i].hash == 0){
	      x[k++] = &(file1_hashes[i+1]);
	    }
	  }
	  x[k++] = &(file2_hashes[0]);
	  for(i = 1;i< file2_hashlens - 1;i++){
	    if(file2_hashes[i].hash == 0){
	      x[k++] = &(file2_hashes[i+1]);
	    }
	  }
	  svm_prob.x = x;
	  #endif
	  if(yimin_trace){
	    for(i = 0;i< k;i++){
	      fprintf(stderr, "\nx[%d]=%lud\n",i,x[i][1].hash);
	    }       
	  }
	  Q_init();
	  solve(); //result is in solver
	  b = calc_b();
	  if(yimin_trace){
	    fprintf(stderr, "b=%lf\n",b);
	  }
	  alpha = (double *)malloc((k1 + k2) * sizeof(double));
	  for(i = 0; i < svm_prob.l; i++)
	    alpha[i] = solver.alpha[i];
	  //free cache
	  cache_free(&svmcache);
	  //compute A,B for sigmoid prediction
	  double *deci_array = (double*) malloc(svm_prob.l * sizeof(double));
	  for(i = 0; i < svm_prob.l; i++){
	    deci_array[i] = calc_decision(svm_prob.x[i], alpha, b);
	  }
	  calc_AB(AB, deci_array, k1, k2);
	  //decision = sigmoid_predict(decision, AB[0], AB[1]);
	  //free(hashes);
	  free(deci_array);
	  
	  fprintf(stderr, "Recalculation of svm hyperplane is finished!\n");
	}else{
	  fprintf(stderr, "There hasn't enough documents to recalculate a svm hyperplane!\n");
	  return (0);
	}	
      }else{

	if(yimin_trace){
	  for(i=0;i<svm_prob.l;i++){
	    j = 0;
	    do{fprintf(stderr, "x[%d][%d]=%lud\n",i,j,svm_prob.x[i][j].hash);}while(svm_prob.x[i][j++].hash!=0);
	  }
	}
	alpha = (double *)malloc((k1 + k2) * sizeof(double));
	for(i = 0; i<(k1 + k2); i++){
	  fread(&alpha[i], sizeof(double), 1, hashf);
	  //fprintf(stderr, "alpha[%d]=%lf\n",i,alpha[i]);      
	}
	fread(&b, sizeof(double), 1, hashf);
	fread(&AB[0], sizeof(double), 1, hashf);
	fread(&AB[1], sizeof(double), 1, hashf);
	fclose(hashf);
	//fprintf(stderr, "b=%lf\n",b);
      }
      decision = calc_decision(hashes,alpha,b);
      //double *deci_array = (double*) malloc(svm_prob.l * sizeof(double));
      //for(i = 0; i < svm_prob.l; i++){
      //deci_array[i] = calc_decision(svm_prob.x[i], alpha, b);
      //}
      //double AB[2];
      //calc_AB(AB, deci_array, k1, k2);
      decision = sigmoid_predict(decision, AB[0], AB[1]);
      free(hashes);
      //free(deci_array);
      
    }//end (k1==0 && k2 ==0)
  }//end (k==0)
  else
    {
      fprintf(stderr, "You need to input (file1.svm | file2.svm| f1vsf2.svmhyp)\n");
    };
  if(svlen > 0){
    char buf [4096];
    buf [0] = '\000';
    
    //   put in standard CRM114 result standard header:
    
    if(decision >= 0.5)
      {
	double pr;
	pr = 10 * log10 (decision + 1e-300) - log10 (1.0 - decision +1e-300 );
	sprintf(buf, 
		"CLASSIFY succeeds; success probability: %6.4f  pR: %6.4f\n", 
		decision, pr);
      }
    else 
      {      
	double pr;
	pr = 10 * log10 (decision + 1e-300) - log10 (1.0 - decision +1e-300 );
	sprintf(buf, 
		"CLASSIFY fails; success probability: %6.4f  pR: %6.4f\n", 
		decision, pr);
      }
    if (strlen (stext) + strlen(buf) <= stext_maxlen)
      strcat (stext, buf);
    
    //   Second line of the status report is the "best match" line:
    //
    //    Here's a template to use
    /*    sprintf (buf, "Best match to file #%ld (%s) prob: %6.4f  pR: %6.4f  \n",
	     bestseen,
	     hashname[bestseen],
	     ptc[bestseen],
	     10 * (log10(ptc[bestseen]) - log10(remainder))
    */

    //   finally, save the status output
    //
    crm_destructive_alter_nvariable (svrbl, svlen, 
				     stext, strlen (stext));
  }
  
  //    Return with the correct status, so an actual FAIL or not can occur.
  if (decision >= 0.5 )
    {
      //   all done... if we got here, we should just continue execution
      if (user_trace)
        fprintf (stderr, "CLASSIFY was a SUCCESS, continuing execution.\n");
    }
  else
    {
      //  Classify was a fail.  Do the skip.
      if (user_trace)
        fprintf (stderr, "CLASSIFY was a FAIL, skipping forward.\n");
      //    and do what we do for a FAIL here
      csl->cstmt = csl->mct[csl->cstmt]->fail_index - 1;
      csl->aliusstk [csl->mct[csl->cstmt]->nest_level] = -1;
      return (0);
    }
 regcomp_failed:
  return (0); 
  
}

#ifdef yimin
//swap two documents in CACHE
static void cache_swap(CACHE *svmcache, int i, int j){
  if(i == j)
    return;
  CACHE_NODE *headi , *headj;
  headi = svmcache->head + i;
  headj = svmcache->head + j;
  if(headi->len) lru_delete(headi);
  if(headj->len) lru_delete(headj);
  float *temp = headi->data;
  headi->data = headj->data;
  headj->data = temp;
  int templen = headi->len;
  headi->len = headj->len;
  headj->len = templen;
  if(headi->len) lru_insert(headi, svmcache);
  if(headj->len) lru_insert(headj, svmcache);
  if(i > j){
    templen = i;
    i = j;
    j = templen;
  }
  CACHE_NODE *tempnode;
  for(tempnode = (svmcache->lru_headnode).next; tempnode != &(svmcache->lru_headnode); tempnode = tempnode->next ){
    if(tempnode->len > i){
      if(tempnode->len > j){
	float tempfloat = tempnode->data[i];
	tempnode->data[i] = tempnode->data[j];
	tempnode->data[j] = tempfloat;
      }else{
	lru_delete(tempnode);
	free(tempnode->data);
	svmcache->size += tempnode->len;
	tempnode->data = 0;
	tempnode->len = 0;
      }
    }      
  }
}



//swap two documents. This includes swap CACHE, svm_prob.x, svm_prob.y and DiagQ
static void swap(int i, int j){
  if(i != j){
    if(1){
      HYPERSPACE_FEATUREBUCKET_STRUCT *temp = svm_prob.x[i];
      svm_prob.x[i] = svm_prob.x[j];
      svm_prob.x[j] = temp;
    }
    if(1){
      float temp = svm_prob.y[i];
      svm_prob.y[i] = svm_prob.y[j];
      svm_prob.y[j] = temp;
    }
    if(1){
      float temp = DiagQ[i];
      DiagQ[i] = DiagQ[j];
      DiagQ[j] = temp;
    }
    cache_swap(&svmcache, i, j);
  }
}

#endif
